#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <ctype.h>

#include "common.h"
#include "macros.h"
#include "buffer.h"
#include "misc.h"
#include "hal-proto.h"
#include "hal-cmd-parser.h"

/* Uncomment it when debug. */
//#define DEBUG_HAL_CMD_PARSER

#define HAL_CMD_CALLBACK(name, arg...) \
do { \
	if (cmd_callbacks.name##_callback != NULL) \
		cmd_callbacks.name##_callback(arg); \
	else \
		cmd_callbacks.not_handled_callback(); \
} while (0)

#define ASCII_LF		0xa

/*
 * The HAL protocol is the text base protocol. It means that all fields are
 * strings. Strings are separeted from each other by an ASCII_LF symbol.
 *
 * Bellow we describe the HAL packet format. The whole packet can't be bigger
 * than HAL_PACKET_MAX bytes. The packet consists of two parts:
 *
 * 	1. A packet header.
 * 	2. A payload part.
 *
 * The header and the payload parts are seprated by extra '\n' symbol.
 * So combination "\n\n" means the end of the header part.
 *
 * The header occupies as max as HAL_HEADER_MAX bytes. The header consists of a
 * protocol magic ("HAL_PROTO"), a command, a subcommand and a length of a
 * payload part.
 *
 * A payload part occupies as max as HAL_PAYLOAD_MAX bytes but actually its
 * size is described by the header. 
 *
 *	+---------------+	<---------------+ HAL_PACKET_MAX
 *	|  PROTO_MAGIC	|	    		|
 *	+---------------+			|
 *	| COMMAND TYPE  |			|
 *	+---------------+	HAL_HEADER_MAX	|
 *	| SUBCOMM. TYPE	|			|
 *	+---------------+			|
 *	| PAYLOAD LENGTH|			|
 *	+---------------+	<---------------|
 *	|     ....	|			|
 *	|    PAYLOAD	|			|
 *	|     ....	|			|
 *	+---------------+	<---------------+
 */
#define HAL_CMD_PARSERSN	26

#ifdef DEBUG_HAL_CMD_PARSER
#define debug(fmt, arg...) fprintf(stderr, "debug: %s: "fmt CRLF, \
						__FUNCTION__, ##arg);
#else
#define debug(fmt, arg...) do {} while(0)
#endif

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
						__FUNCTION__,  ##arg);

typedef void (*hal_cmd_parser_t)(struct buffer *pkt, void *opaque);

struct hal_cmd_parser {
	const char		*cmd;
	const char		*subcmd;
	hal_cmd_parser_t	handler;
};

static struct hal_cmd_callbacks cmd_callbacks;
static struct hal_cmd_parser parsers[HAL_CMD_PARSERSN];

/* Below you see protocol commands and statuses. The protocol is
 * text-based, it means that we often have to compare strings. We declare
 * all protocol strings as constant variables. Indeed our variables are
 * constant pointers that point to constant strings.
 *
 * In our implementation it helps us to reduce the number of string comparison
 * operations. We use fast look up functions to detect the types of commands
 * and statuses and then we work with the returned constant pointer as with
 * a common BINARY value. In the current implementation we call the STREQ
 * macros only once for each packet to match the protocol magic.
 */

/* List of commands, subcommands and statuses we can parse. */
const char const *hal_cmd_connect = "CN";
const char const *hal_cmd_disconnect = "DC";
const char const *hal_cmd_cycle_begin = "CB";
const char const *hal_cmd_cycle_run = "CR";
const char const *hal_cmd_cycle_end = "CE";
const char const *hal_cmd_write_inputs = "WI";
const char const *hal_cmd_write_outputs = "WO";
const char const *hal_cmd_write_end = "WE";
const char const *hal_cmd_write_prepare= "WP";
const char const *hal_cmd_read_inputs = "RI";
const char const *hal_cmd_read_outputs = "RO";
const char const *hal_cmd_read_end = "RE";
const char const *hal_cmd_read_prepare = "RP";

const char const *hal_subcmd_request = "REQ";
const char const *hal_subcmd_reply = "REP";

const char const *hal_status_ok = "OK";
const char const *hal_status_error = "ERR";

static const char const *str_to_cmd(const char *str)
{
	const char *p;

	p = str;
	switch (p[0]) {
	case 'C':
		switch (p[1]) {
		case 'N':
			return HAL_CMD_CONNECT;
		case 'B':
			return HAL_CMD_CYCLE_BEGIN;
		case 'R':
			return HAL_CMD_CYCLE_RUN;
		case 'E':
			return HAL_CMD_CYCLE_END;
		}
		break;
	case 'W':
		switch (p[1]) {
		case 'I':
			return HAL_CMD_WRITE_INPUTS;
		case 'O':
			return HAL_CMD_WRITE_OUTPUTS;
		case 'P':
			return HAL_CMD_WRITE_PREPARE;
		case 'E':
			return HAL_CMD_WRITE_END;
		}
		break;
	case 'R':
		switch (p[1]) {
		case 'I':
			return HAL_CMD_READ_INPUTS;
		case 'O':
			return HAL_CMD_READ_OUTPUTS;
		case 'P':
			return HAL_CMD_READ_PREPARE;
		case 'E':
			return HAL_CMD_READ_END;
		}
		break;
	case 'D':
		switch (p[1]) {
		case 'C':
			return HAL_CMD_DISCONNECT;
		}
		break;
	}
	
	return NULL;
}

static const char const *str_to_status(const char *str)
{
	const char *p;

	p = str;
	switch (p[0]) {
	case 'O':
		switch (p[1]) {
		case 'K':
			return HAL_OK;
		}
		break;
	case 'E':
		if (p[1] == 'R' && p[2] == 'R')
			return HAL_ERR;
		break;
	}

	return NULL;
}

static const char const *str_to_subcmd(const char *str)
{
	const char *p;

	p = str;

	if (p[0] != 'R' || p[1] != 'E')
		return NULL;

	switch (p[2]) {
	case 'Q':
		return HAL_SUBCMD_REQUEST;
	case 'P':
		return HAL_SUBCMD_REPLY;
	}

	return NULL;
}
/* Simple wrapper for extracting an unsigned integer number. */
static int hal_cmd_get_arg_uint(struct buffer *pkt, unsigned int *num)
{
	char sbuf[HAL_STRING_MAX];
	size_t slen;
	char *str;
	int ret;

	str = extract_string(pkt, sbuf, sizeof(sbuf), &slen, ASCII_LF);
	if (str == NULL)
		return -1;

	buffer_consume(pkt, slen+1);

	ret = str_to_uint(str, NULL, num, 16);
	if (ret < 0)
		return -1;

	return 0;
}

/* Simple wrapper for extracting a string. */
static int hal_cmd_get_arg_str(struct buffer *pkt, char *s, size_t max_len)
{
	char *str;
	size_t slen;

	str = extract_string(pkt, s, max_len, &slen, ASCII_LF);
	if (str == NULL)
		return -1;

	buffer_consume(pkt, slen+1);

	return 0;
}

/* Connect request has two arguments -- major and minor versions of protocol. */
static void hal_cmd_connect_request(struct buffer *pkt, void *opaque)
{
	unsigned int major, minor;
	int ret;

	ret = hal_cmd_get_arg_uint(pkt, &major);
	if (ret < 0) { 
		warn("CN request: major version not found");
		return;
	}

	ret = hal_cmd_get_arg_uint(pkt, &minor);
	if (ret < 0) {
		warn("CN request: minor version not found");
		return;
	}

	if (major != HAL_PROTO_MAJOR || minor != HAL_PROTO_MINOR) {
		warn("CN request: proto version not matched %d.%d",
							major, minor);
		return;
	}

	debug("CN request: OK major %d minor %d", major, minor);

	HAL_CMD_CALLBACK(connect_request, opaque);
}

/* Wrapper for commands that do not have arguments. We call only the right
 * callback. */
static void hal_cmd_noargs_request(const char *cmd, void *opaque)
{
	/* Call request callbacks according to a command. */
	if (cmd == HAL_CMD_DISCONNECT)
		HAL_CMD_CALLBACK(disconnect_request, opaque);
	else if (cmd == HAL_CMD_CYCLE_BEGIN)
		HAL_CMD_CALLBACK(cycle_begin_request, opaque);
	else if (cmd == HAL_CMD_CYCLE_RUN)
		HAL_CMD_CALLBACK(cycle_run_request, opaque);
	else if (cmd == HAL_CMD_CYCLE_END)
		HAL_CMD_CALLBACK(cycle_end_request, opaque);
	else if (cmd == HAL_CMD_WRITE_PREPARE)
		HAL_CMD_CALLBACK(write_prepare_request, opaque);
	else if (cmd == HAL_CMD_WRITE_END)
		HAL_CMD_CALLBACK(write_end_request, opaque);
	else if (cmd == HAL_CMD_READ_PREPARE)
		HAL_CMD_CALLBACK(read_prepare_request, opaque);
	else if (cmd == HAL_CMD_READ_END)
		HAL_CMD_CALLBACK(read_end_request, opaque);
}
/* Below you see handlers for request commands with no arguments. The 
 * payload length is zero. */
static void hal_cmd_disconnect_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_request(HAL_CMD_DISCONNECT, opaque);
}

static void hal_cmd_cycle_begin_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_request(HAL_CMD_CYCLE_BEGIN, opaque);
}

static void hal_cmd_cycle_run_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_request(HAL_CMD_CYCLE_RUN, opaque);
}

static void hal_cmd_cycle_end_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_request(HAL_CMD_CYCLE_END, opaque);
}

static void hal_cmd_write_prepare_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_request(HAL_CMD_WRITE_PREPARE, opaque);
}

static void hal_cmd_write_end_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_request(HAL_CMD_WRITE_END, opaque);
}

static void hal_cmd_read_prepare_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_request(HAL_CMD_READ_PREPARE, opaque);
}

static void hal_cmd_read_end_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_request(HAL_CMD_READ_END, opaque);
}
/* Generic handler for a write request commands. First extract length and
 * offset arguments then get an image. Finnaly call the right callback. */
static void hal_cmd_write_image_request(const char *cmd, struct buffer *pkt,
					void *opaque)
{
	char image[HAL_IMAGE_MAX+1];
	unsigned int off;
	int ret;

	ret = hal_cmd_get_arg_uint(pkt, &off);
	if (ret < 0) {
		warn("%s: offset not found", cmd);
		return;
	}

	/* Time to get the image that a user want to write. */
	ret = hal_cmd_get_arg_str(pkt, image, sizeof(image));
	if (ret < 0) {
		warn("%s: image not found", cmd);
		return;
	}

	debug("%s request: offset %d image %s", cmd, off, image);

	/* Call write inputs or outputs callbacks. */
	if (cmd == HAL_CMD_WRITE_INPUTS)
		HAL_CMD_CALLBACK(write_inputs_request, image, off, opaque);
	else if (cmd == HAL_CMD_WRITE_OUTPUTS)
		HAL_CMD_CALLBACK(write_outputs_request, image, off, opaque);
}

static void hal_cmd_write_inputs_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_write_image_request(HAL_CMD_WRITE_INPUTS, pkt, opaque);
}

static void hal_cmd_write_outputs_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_write_image_request(HAL_CMD_WRITE_OUTPUTS, pkt, opaque);
}
/* Generic handler for a read request commands. We just extract length and
 * offset arguments. */
static void hal_cmd_read_image_request(const char *cmd, struct buffer *pkt,
					void *opaque)
{
	unsigned int len, off;
	int ret;
	
	ret = hal_cmd_get_arg_uint(pkt, &len);
	if (ret < 0) {
		warn("%s: length not found", cmd);
		return;
	}

	ret = hal_cmd_get_arg_uint(pkt, &off);
	if (ret < 0) {
		warn("%s: offset not found", cmd);
		return;
	}

	debug("%s request: length %d offset %d", cmd, len, off);

	/* Call read inputs or outputs callbacks. */
	if (cmd == HAL_CMD_READ_INPUTS)
		HAL_CMD_CALLBACK(read_inputs_request, len, off, opaque);
	else if (cmd == HAL_CMD_READ_OUTPUTS)
		HAL_CMD_CALLBACK(read_outputs_request, len, off, opaque);
}

static void hal_cmd_read_inputs_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_read_image_request(HAL_CMD_READ_INPUTS, pkt, opaque);
}

static void hal_cmd_read_outputs_request(struct buffer *pkt, void *opaque)
{
	hal_cmd_read_image_request(HAL_CMD_READ_OUTPUTS, pkt, opaque);
}

/* The first reply argument in the payload is the request status. Command
 * request can be successfull (HAL_OK) or not (HAL_ERR). */
static const char *hal_cmd_get_reply_status(const char *cmd, struct buffer *pkt)
{
	char sbuf[HAL_STRING_MAX];
	const char *status;
	int ret;

	ret = hal_cmd_get_arg_str(pkt, sbuf, sizeof(sbuf));
	if (ret < 0) {
		warn("%s: reply status not found", cmd);
		return NULL;
	} 

	status = str_to_status(sbuf);

	return status;
}

/* Parse a connect reply command. First get a command status if it was not
 * success then get a custom user error message, otherwise process command
 * arguments -- inputs and outputs images sizes of other side. */
static void hal_cmd_connect_reply(struct buffer *pkt, void *opaque)
{
	const char *status;
	char *error;
	char ebuf[HAL_STRING_MAX];
	char side[2];
	unsigned int inputs_sz, outputs_sz;
	int ret;

	error = NULL;

	status = hal_cmd_get_reply_status(HAL_CMD_CONNECT, pkt);
	if (status == NULL)
		return;

	if (status == HAL_ERR) {
		ret = hal_cmd_get_arg_str(pkt, ebuf, sizeof(ebuf));
		if (ret < 0) {
			warn("CN reply: error message not found");
			return;
		}
		error = ebuf;
		goto call;
	}
	
	ret = hal_cmd_get_arg_uint(pkt, &inputs_sz);
	if (ret < 0) {
		warn("CN reply: inputs size not found");
		return;
	}

	ret = hal_cmd_get_arg_uint(pkt, &outputs_sz);
	if (ret < 0) {
		warn("CN reply: outputs size not found");
		return;
	}

	ret = hal_cmd_get_arg_str(pkt, side, sizeof(side));
	if (ret < 0) {
		warn("CN reply: side argument not found");
		return;
	}

	switch (side[0]) {
	case 'S':
	case 's':
	case 'M':
	case 'm':
		break;
	default:
		warn("CN reply: wrong side argument %c", side[0]);
		return;
	}

	debug("CN reply: OK inputs size %d outputs size %d", inputs_sz, outputs_sz);

call:
	HAL_CMD_CALLBACK(connect_reply, inputs_sz, outputs_sz, tolower(side[0]),  error, opaque);
}

/* Generic handler for reply commands where we just need to get a status. If
 * it was not success we extract a user custom error message. */
static void hal_cmd_noargs_reply(const char *cmd, struct buffer *pkt,
				 void *opaque)
{
	const char *status;
	char *error;
	char ebuf[HAL_STRING_MAX];
	int ret;

	error = NULL;

	status = hal_cmd_get_reply_status(cmd, pkt);
	if (status == NULL)
		return;

	if (status == HAL_ERR) {
		ret = hal_cmd_get_arg_str(pkt, ebuf, sizeof(ebuf));
		if (ret < 0) {
			warn("%s reply: error message not found", cmd);
			return;
		}
		error = ebuf;
	}
	/* Call reply callbacks according to a command. */
	if (cmd == HAL_CMD_DISCONNECT)
		HAL_CMD_CALLBACK(disconnect_reply, error, opaque);
	else if (cmd == HAL_CMD_CYCLE_BEGIN)
		HAL_CMD_CALLBACK(cycle_begin_reply, error, opaque);
	else if (cmd == HAL_CMD_CYCLE_RUN)
		HAL_CMD_CALLBACK(cycle_run_reply, error, opaque);
	else if (cmd == HAL_CMD_CYCLE_END)
		HAL_CMD_CALLBACK(cycle_end_reply, error, opaque);
	else if (cmd == HAL_CMD_WRITE_INPUTS)
		HAL_CMD_CALLBACK(write_inputs_reply, error, opaque);
	else if (cmd == HAL_CMD_WRITE_OUTPUTS)
		HAL_CMD_CALLBACK(write_outputs_reply, error, opaque);
	else if (cmd == HAL_CMD_WRITE_PREPARE)
		HAL_CMD_CALLBACK(write_prepare_reply, error, opaque);
	else if (cmd == HAL_CMD_WRITE_END)
		HAL_CMD_CALLBACK(write_end_reply, error, opaque);
	else if (cmd == HAL_CMD_READ_PREPARE)
		HAL_CMD_CALLBACK(read_prepare_reply, error, opaque);
	else if (cmd == HAL_CMD_READ_END)
		HAL_CMD_CALLBACK(read_end_reply, error, opaque);
}
/* Below you see handlers where payload consists of only the status. */
static void hal_cmd_disconnect_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_DISCONNECT, pkt, opaque);
}

static void hal_cmd_cycle_begin_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_CYCLE_BEGIN, pkt, opaque);
}

static void hal_cmd_cycle_run_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_CYCLE_RUN, pkt, opaque);
}

static void hal_cmd_cycle_end_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_CYCLE_END, pkt, opaque);
}

static void hal_cmd_write_inputs_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_WRITE_INPUTS, pkt, opaque);
}

static void hal_cmd_write_outputs_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_WRITE_OUTPUTS, pkt, opaque);
}

static void hal_cmd_write_prepare_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_WRITE_PREPARE, pkt, opaque);
}

static void hal_cmd_write_end_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_WRITE_END, pkt, opaque);
}

static void hal_cmd_read_prepare_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_READ_PREPARE, pkt, opaque);
}

static void hal_cmd_read_end_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_noargs_reply(HAL_CMD_READ_END, pkt, opaque);
}
/* Generic handler for read reply commands where we get a status then if it
 * was not success extract a user custom error message, otherwise get lenght
 * and offset arguments and the read image. */
static void hal_cmd_read_image_reply(const char *cmd, struct buffer *pkt,
				     void *opaque)
{
	const char *status;
	char *error;
	char ebuf[HAL_STRING_MAX];
	char image[HAL_IMAGE_MAX+1];
	unsigned int off;
	int ret;

	error = NULL;

	status = hal_cmd_get_reply_status(cmd, pkt);
	if (status == NULL)
		return;

	if (status == HAL_ERR) {
		ret = hal_cmd_get_arg_str(pkt, ebuf, sizeof(ebuf));
		if (ret < 0) {
			warn("%s reply: error message not found", cmd);
			return;
		}
		error = ebuf;
		goto call;
	}

	ret = hal_cmd_get_arg_uint(pkt, &off);
	if (ret < 0) {
		warn("%s: offset not found", cmd);
		return;
	}
	
	ret = hal_cmd_get_arg_str(pkt, image, sizeof(image));
	if (ret < 0) {
		warn("%s: image not found", cmd);
		return;
	}

	debug("%s reply: offset %d image %s", cmd, off, image);
call:

	if (cmd == HAL_CMD_READ_INPUTS)
		HAL_CMD_CALLBACK(read_inputs_reply, image, off, error, opaque);
	else if (cmd == HAL_CMD_READ_OUTPUTS)
		HAL_CMD_CALLBACK(read_outputs_reply, image, off, error, opaque);
}

static void hal_cmd_read_inputs_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_read_image_reply(HAL_CMD_READ_INPUTS, pkt, opaque);
}

static void hal_cmd_read_outputs_reply(struct buffer *pkt, void *opaque)
{
	hal_cmd_read_image_reply(HAL_CMD_READ_OUTPUTS, pkt, opaque);
}

/* Parse a payload data. */
static void parse_payload(struct buffer *pkt, const char *cmd,
			  const char *subcmd, void *opaque)
{
	unsigned int i;

	buffer_putc(pkt, 0);

	for (i = 0; i < HAL_CMD_PARSERSN; i++) {
		if (parsers[i].cmd == cmd && parsers[i].subcmd == subcmd) {
			debug("call handler for %s.%s", cmd, subcmd);
			parsers[i].handler(pkt, opaque);
			buffer_trim(pkt, 1);
			return;	
		}
	}

	warn("unknown command %s.%s", cmd, subcmd);
}
/* A header and a payload are separated by two '\n'. Also calculate the header
 * length */
static int header_present(struct buffer *buf, unsigned int *hlen)
{
	char *beg, *end;
	int ret;

	buffer_putc(buf, 0);
	beg = buffer_ptr(buf);

	if ((end = strstr(beg, "\n\n")) != NULL) {
		ret = 1;
		*hlen = end-beg+2;
	} else {
		ret = 0;
	}

	buffer_trim(buf, 1);

	return ret;
}

static int parse_header(struct buffer *buf, const char **cmd,
			const char **subcmd, unsigned int *plen)
{
	char sbuf[HAL_STRING_MAX];
	int ret;

	buffer_putc(buf, 0);

	/* Get a protocol magic. */
	ret = hal_cmd_get_arg_str(buf, sbuf, sizeof(sbuf));
	if (ret < 0) {
		warn("magic not found");
		goto out;
	}

	if (!STREQ(sbuf, HAL_PROTO_MAGIC)) {
		warn("magic not matched %s", sbuf);
		goto out;
	}
	/* Get a command. */	
	ret = hal_cmd_get_arg_str(buf, sbuf, sizeof(sbuf));
	if (ret < 0) {
		warn("command name not found");
		goto out;
	}

	*cmd = str_to_cmd(sbuf);
	if (*cmd == NULL) {
		warn("unknown command %s", sbuf);
		goto out;
	}
	/* Get a subcommand. */
	ret = hal_cmd_get_arg_str(buf, sbuf, sizeof(sbuf));
	if (ret < 0) {
		warn("subcommand name not found");
		goto out;
	}

	*subcmd = str_to_subcmd(sbuf);
	if (*subcmd == NULL) {
		warn("unknown subcommand %s", sbuf);
		goto out;
	}
	/* Get the payload length. */
	ret = hal_cmd_get_arg_uint(buf, plen);
	if (ret < 0) {
		warn("packet length not found");
		goto out;	
	}
	/* Eat the header end. */
	buffer_consume(buf, 1);
	buffer_trim(buf, 1);

	return 0;
out:
	buffer_trim(buf, 1);
	return -1;
}
/* The main routine, in the cycle we obtain packets, then try to find a header,
 * then parse the header (protocol magic, payload length) and the payload. */
void hal_cmd_parse_packet(struct buffer *buf, void *opaque)
{
	char pbuf[HAL_PAYLOAD_MAX+1];
	char hbuf[HAL_HEADER_MAX+1];
	struct buffer *payload;
	struct buffer *header;
	unsigned int payload_len, header_len, rest_len;
	const char *cmd, *subcmd;
	int res;

	header = buffer_create_static(hbuf, sizeof(hbuf));
	payload = buffer_create_static(pbuf, sizeof(pbuf));
	/* At least header must be presented. */
	while (header_present(buf, &header_len)) {
		if (header_len > sizeof(hbuf)) {
			warn("header length is large");
			buffer_reset(buf);
			break;
		}
		/* Read the header from the main buffer and parse it */
		buffer_copy(header, buf, header_len);
		res = parse_header(header, &cmd, &subcmd, &payload_len);
		if (res < 0) {
			buffer_reset(buf);
			break;
		}

		buffer_reset(header);
		
		rest_len = buffer_size(buf) - header_len;
		debug("rest buffer length %d payload length  %d", rest_len, payload_len);
		debug("found cmd %s and subcmd %s", cmd, subcmd);
		/* A buffer can hold the last payload data or some more packets,
		 * also a payload length can't be bigger than our payload
		 * length limit. */
		if (rest_len >= payload_len && payload_len <= sizeof(pbuf)) {
			/* Only now we can consume the header. Then parse the payload. */
			buffer_consume(buf, header_len);
			buffer_copy(payload, buf, payload_len);
			buffer_consume(buf, payload_len);
			parse_payload(payload, cmd, subcmd, opaque);
			buffer_reset(payload);
		} else if (payload_len > sizeof(pbuf)) {
			warn("payload length is large");
			buffer_reset(buf);
			break;
		} else if (rest_len < payload_len) {
			/* The packet is not full, need to read more. We don't
			 * consume the header it is still in the main buffer. */
			warn("payload length is too small against buffer size: %d %d",
						rest_len, payload_len);
			break;
		}
	}
	buffer_free(payload);
	buffer_free(header);
}

static int parser_init()
{
	unsigned int i;

	struct hal_cmd_parser parsers_tmp[] = {
		{ HAL_CMD_CONNECT,	HAL_SUBCMD_REQUEST,  hal_cmd_connect_request },
		{ HAL_CMD_CONNECT,	HAL_SUBCMD_REPLY,    hal_cmd_connect_reply },
		{ HAL_CMD_DISCONNECT,	HAL_SUBCMD_REQUEST,  hal_cmd_disconnect_request },
		{ HAL_CMD_DISCONNECT,	HAL_SUBCMD_REPLY,    hal_cmd_disconnect_reply },
		{ HAL_CMD_CYCLE_BEGIN,	HAL_SUBCMD_REQUEST,  hal_cmd_cycle_begin_request },
		{ HAL_CMD_CYCLE_BEGIN,	HAL_SUBCMD_REPLY,    hal_cmd_cycle_begin_reply },
		{ HAL_CMD_CYCLE_RUN,	HAL_SUBCMD_REQUEST,  hal_cmd_cycle_run_request },
		{ HAL_CMD_CYCLE_RUN,	HAL_SUBCMD_REPLY,    hal_cmd_cycle_run_reply },
		{ HAL_CMD_CYCLE_END,	HAL_SUBCMD_REQUEST,  hal_cmd_cycle_end_request },
		{ HAL_CMD_CYCLE_END,	HAL_SUBCMD_REPLY,    hal_cmd_cycle_end_reply },
		{ HAL_CMD_WRITE_INPUTS, HAL_SUBCMD_REQUEST,  hal_cmd_write_inputs_request },
		{ HAL_CMD_WRITE_INPUTS, HAL_SUBCMD_REPLY,    hal_cmd_write_inputs_reply },
		{ HAL_CMD_WRITE_OUTPUTS, HAL_SUBCMD_REQUEST, hal_cmd_write_outputs_request },
		{ HAL_CMD_WRITE_OUTPUTS, HAL_SUBCMD_REPLY,   hal_cmd_write_outputs_reply },
		{ HAL_CMD_WRITE_PREPARE, HAL_SUBCMD_REQUEST, hal_cmd_write_prepare_request },
		{ HAL_CMD_WRITE_PREPARE, HAL_SUBCMD_REPLY,   hal_cmd_write_prepare_reply },
		{ HAL_CMD_WRITE_END,	HAL_SUBCMD_REQUEST,  hal_cmd_write_end_request },
		{ HAL_CMD_WRITE_END,	HAL_SUBCMD_REPLY,    hal_cmd_write_end_reply },
		{ HAL_CMD_READ_INPUTS,	HAL_SUBCMD_REQUEST,  hal_cmd_read_inputs_request },
		{ HAL_CMD_READ_INPUTS,	HAL_SUBCMD_REPLY,    hal_cmd_read_inputs_reply },
		{ HAL_CMD_READ_OUTPUTS,	HAL_SUBCMD_REQUEST,  hal_cmd_read_outputs_request },
		{ HAL_CMD_READ_OUTPUTS,	HAL_SUBCMD_REPLY,    hal_cmd_read_outputs_reply },
		{ HAL_CMD_READ_PREPARE, HAL_SUBCMD_REQUEST,  hal_cmd_read_prepare_request },
		{ HAL_CMD_READ_PREPARE, HAL_SUBCMD_REPLY,    hal_cmd_read_prepare_reply },
		{ HAL_CMD_READ_END,	HAL_SUBCMD_REQUEST,  hal_cmd_read_end_request },
		{ HAL_CMD_READ_END,	HAL_SUBCMD_REPLY,    hal_cmd_read_end_reply }
	};

	for (i = 0; i < HAL_CMD_PARSERSN; i++)
		parsers[i] = parsers_tmp[i];

	return 0;
}

int hal_cmd_parser_register_callbacks(struct hal_cmd_callbacks *new_cb)
{
	static int init = 0;

	return_val_if_fail(new_cb != NULL, -1);

	if (!init) {
		parser_init();
		init = 1;
	}

	cmd_callbacks = *new_cb;

	return 0;
}

